package com.nokia.dingtalk_api.service;

import com.aliyun.dingtalkrobot_1_0.models.BatchSendOTOResponse;
import com.aliyun.dingtalkrobot_1_0.models.OrgGroupSendResponse;
import com.baomidou.mybatisplus.core.conditions.query.QueryWrapper;
import com.baomidou.mybatisplus.core.metadata.OrderItem;
import com.baomidou.mybatisplus.extension.plugins.pagination.Page;
import com.dingtalk.api.response.OapiMediaUploadResponse;
import com.google.gson.Gson;
import com.nokia.dingtalk_api.common.R;
import com.nokia.dingtalk_api.common.exception.BizException;
import com.nokia.dingtalk_api.common.exception.MyRuntimeException;
import com.nokia.dingtalk_api.config.AppConfig;
import com.nokia.dingtalk_api.dao.IAppTaskLogService;
import com.nokia.dingtalk_api.dao.IAppTaskService;
import com.nokia.dingtalk_api.mapper.AppTaskMapper;
import com.nokia.dingtalk_api.pojos.bo.AppTaskBo;
import com.nokia.dingtalk_api.pojos.dto.cms.AddAppTaskDto;
import com.nokia.dingtalk_api.pojos.dto.cms.DeleteAppTaskDto;
import com.nokia.dingtalk_api.pojos.dto.cms.EnableAppTaskDto;
import com.nokia.dingtalk_api.pojos.dto.cms.ListAppTaskDto;
import com.nokia.dingtalk_api.pojos.dto.cms.ListAppTaskLogDto;
import com.nokia.dingtalk_api.pojos.dto.cms.RunAppTaskDto;
import com.nokia.dingtalk_api.pojos.dto.cms.StopAppTaskDto;
import com.nokia.dingtalk_api.pojos.dto.cms.UpdateAppTaskDto;
import com.nokia.dingtalk_api.pojos.enums.AppTaskStatusEnum;
import com.nokia.dingtalk_api.pojos.enums.SortEnum;
import com.nokia.dingtalk_api.pojos.po.AppTaskLogPo;
import com.nokia.dingtalk_api.pojos.po.AppTaskPo;
import com.nokia.dingtalk_api.pojos.vo.PageVo;
import com.nokia.dingtalk_api.pojos.vo.cms.ListAppTaskVo;
import com.nokia.dingtalk_api.util.DingTalkApiUtil;
import com.taobao.api.FileItem;
import jakarta.annotation.PostConstruct;
import lombok.Data;
import lombok.RequiredArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.apache.sshd.sftp.client.SftpClient;
import org.springframework.integration.file.remote.SessionCallbackWithoutResult;
import org.springframework.integration.sftp.session.DefaultSftpSessionFactory;
import org.springframework.integration.sftp.session.SftpRemoteFileTemplate;
import org.springframework.scheduling.concurrent.ThreadPoolTaskScheduler;
import org.springframework.scheduling.support.CronTrigger;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.util.CollectionUtils;
import org.springframework.util.StringUtils;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.nio.file.attribute.AclEntry;
import java.nio.file.attribute.FileTime;
import java.util.Arrays;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Optional;
import java.util.Set;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.ConcurrentMap;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.TimeUnit;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.stream.Collectors;

@Slf4j
@Service
@RequiredArgsConstructor
public class AppTaskService {
    /**
     * 英文逗号分割的手机号列表
     */
    static final Pattern PHONE_LIST_PATTERN = Pattern.compile("\\d{11}(,\\d{11})*");
    /**
     * 英文逗号分割的userId列表
     */
    static final Pattern USER_ID_LIST_PATTERN = Pattern.compile("\\w+(,\\w+)*");
    static final ConcurrentMap<String, ScheduledFuture<?>> TASKS = new ConcurrentHashMap<>();
    private final ThreadPoolTaskScheduler taskScheduler;
    private final AppTaskMapper appTaskMapper;
    private final IAppTaskLogService iAppTaskLogService;
    private final OpenService openService;
    private final AesService aesService;
    private final AppConfig appConfig;
    private final IAppTaskService iAppTaskService;

    /**
     * 删除任务
     * @param id 任务id
     */
    public void removeTask(String id) {
        ScheduledFuture<?> future = TASKS.get(id);
        if (future != null && !future.isCancelled()) {
            future.cancel(false);
        }
        TASKS.remove(id);
    }

    /**
     * 添加任务
     * @param t 任务
     */
    public void addTask(AppTaskBo t) {
        removeTask(t.getTaskId());
        CronTrigger cronTrigger = new CronTrigger(t.getCron());
        ScheduledFuture<?> future = taskScheduler.schedule(() -> runTask(t), cronTrigger);
        TASKS.put(t.getTaskId(), future);
    }

    /**
     * 运行任务
     * @param t 任务
     */
    public void runTask(AppTaskBo t) {
        AppTaskLogPo appTaskLogPo = new AppTaskLogPo(t);
        Gson gson = new Gson();
        try {
            CompletableFuture.runAsync(() -> {
                log.info("开始执行应用任务: {}", appTaskLogPo);
                DefaultSftpSessionFactory factory = new DefaultSftpSessionFactory();
                factory.setHost(t.getHost());
                factory.setPort(t.getPort());
                factory.setUser(t.getUsername());
                factory.setPassword(aesService.decrypt(t.getPassword()));
                factory.setAllowUnknownKeys(true);
                SftpRemoteFileTemplate template = new SftpRemoteFileTemplate(factory);
                // 查找数据目录匹配的最新文件
                Pattern pattern = Pattern.compile(t.getFilePattern());
                Optional<DirEntryBo> dirEntryBoOptional = Arrays.stream(template.list(t.getDataDir()))
                        .filter(tt -> pattern.matcher(tt.getFilename()).find())
                        .map(DirEntryBo::new).max(Comparator.comparing(DirEntryBo::getModifyTime));
                DirEntryBo dirEntryBo = dirEntryBoOptional.orElseThrow(() -> new BizException("没有找到匹配的文件"));
                String filename = dirEntryBo.getFilename();
                // 读取文件
                String filePath = t.getDataDir() + (t.getDataDir().endsWith("/") ? "" : "/") + filename;
                appTaskLogPo.setFilePath(filePath);
                byte[] bytes;
                try (ByteArrayOutputStream os = new ByteArrayOutputStream()) {
                    template.execute((SessionCallbackWithoutResult<SftpClient.DirEntry>) session
                            -> session.read(filePath, os));
                    bytes = os.toByteArray();
                } catch (IOException e) {
                    throw new MyRuntimeException(e);
                }
                // 检查文件大小
                OpenService.checkFileSize(bytes.length);
                String fileExtension = OpenService.getFileExtension(filename);
                String type = OpenService.getUploadType(fileExtension);
                FileItem fileItem = new FileItem(filename, bytes);
                String robotCode = t.getRobotCode();
                String openConversationId = t.getOpenConversationId();
                String accessToken = openService.getAccessToken(robotCode, t.getRobotSecret());
                // 上传文件
                OapiMediaUploadResponse uploadResponse = DingTalkApiUtil.upload(type, fileItem, accessToken);
                String mediaId = uploadResponse.getMediaId();
                // 发送单聊消息
                if (t.getConversationType().equals(1)) {
                    Set<String> phones = new HashSet<>();
                    if (StringUtils.hasText(t.getPhones())) {
                        phones = Arrays.stream(t.getPhones().split(",")).collect(Collectors.toSet());
                    }
                    Set<String> userIds = new HashSet<>();
                    if (StringUtils.hasText(t.getUserIds())) {
                        userIds = Arrays.stream(t.getUserIds().split(",")).collect(Collectors.toSet());
                    }
                    openService.checkPhonesUserIds(phones, userIds);
                    Map<String, String> failPhones = new HashMap<>();
                    List<String> userIdList = openService.getUserIds(userIds, phones, accessToken, failPhones);
                    // 发送图片
                    if ("image".equals(type)) {
                        BatchSendOTOResponse response = DingTalkApiUtil.batchSendOtoSampleImageMsg(mediaId,
                                accessToken, robotCode, userIdList);
                        appTaskLogPo.setDetail(gson.toJson(response));
                    } else {
                        // 发送文件
                        BatchSendOTOResponse response = DingTalkApiUtil.batchSendOtoSampleFile(mediaId, filename,
                                fileExtension, accessToken, robotCode, userIdList);
                        appTaskLogPo.setDetail(gson.toJson(response));
                    }
                } else {
                    // 发送群聊消息
                    if ("image".equals(type)) {
                        // 发送图片
                        OrgGroupSendResponse response = DingTalkApiUtil.groupSendSampleImageMsg(mediaId,
                                accessToken, openConversationId, robotCode);
                        appTaskLogPo.setDetail(gson.toJson(response));
                    } else {
                        // 发送文件
                        OrgGroupSendResponse response = DingTalkApiUtil.groupSendSampleFile(mediaId, filename,
                                fileExtension, accessToken, openConversationId, robotCode);
                        appTaskLogPo.setDetail(gson.toJson(response));
                    }
                }
                log.info("结束应用任务: {}", appTaskLogPo);
            }).get(appConfig.getTaskTimeout(), TimeUnit.MINUTES);
        } catch (InterruptedException e) {
            appTaskLogPo.setStatus(0);
            appTaskLogPo.setDetail(e.toString());
            log.error("线程中断: {}, {}", appTaskLogPo, e, e);
            Thread.currentThread().interrupt();
        } catch (Exception e) {
            appTaskLogPo.setStatus(0);
            appTaskLogPo.setDetail(e.toString());
            log.error("应用任务执行失败: {}, {}", appTaskLogPo, e, e);
        } finally {
            // 记录日志
            iAppTaskLogService.save(appTaskLogPo);
        }
    }

    /**
     * 应用定时任务初始化
     */
    @PostConstruct
    public void init() {
        List<AppTaskBo> list = appTaskMapper.getRunningAppTasks(AppTaskStatusEnum.RUNNING.value);
        list.forEach(this::addTask);
    }

    /**
     * 检查任务参数
     * @param dto 任务参数
     */
    private static void checkAppTaskDto(AddAppTaskDto dto) {
        if (dto.getConversationType().equals(1)) {
            if (!StringUtils.hasText(dto.getPhones()) && !StringUtils.hasText(dto.getUserIds())) {
                throw new BizException("phones或userIds必须有一个不能为空");
            }
            if (StringUtils.hasText(dto.getPhones())) {
                Matcher matcher = PHONE_LIST_PATTERN.matcher(dto.getPhones());
                if (!matcher.matches()) {
                    throw new BizException("phones格式不正确");
                }
            } else {
                dto.setPhones("");
            }
            if (StringUtils.hasText(dto.getUserIds())) {
                Matcher matcher = USER_ID_LIST_PATTERN.matcher(dto.getUserIds());
                if (!matcher.matches()) {
                    throw new BizException("userIds格式不正确");
                }
            } else {
                dto.setUserIds("");
            }
            dto.setOpenConversationId("");
        }
        if (dto.getConversationType().equals(2)) {
            if (!StringUtils.hasText(dto.getOpenConversationId())) {
                throw new BizException("openConversationId不能为空");
            }
            dto.setPhones("");
            dto.setUserIds("");
        }
    }

    public R<PageVo<ListAppTaskVo>> listAppTask(ListAppTaskDto dto) {
        Page<AppTaskBo> page = new Page<>(dto.getCurrent(), dto.getPageSize());
        if (CollectionUtils.isEmpty(dto.getSort())) {
            page.addOrder(OrderItem.desc("update_time"));
        } else {
            List<OrderItem> items = dto.getSort().entrySet().stream()
                    .map(t -> t.getValue().equals(SortEnum.ascend) ? OrderItem.asc(t.getKey().value)
                            : OrderItem.desc(t.getKey().value)).toList();
            page.addOrder(items);
        }
        List<AppTaskBo> list = appTaskMapper.listAppTask(page, dto);
        List<ListAppTaskVo> vos = list.stream().map(ListAppTaskVo::new).toList();
        return R.ok(new PageVo<>(page.getTotal(), vos));
    }

    @Transactional(rollbackFor = Exception.class)
    public R<Object> addAppTask(AddAppTaskDto dto) {
        checkAppTaskDto(dto);
        QueryWrapper<AppTaskPo> wrapper = new QueryWrapper<>();
        wrapper.eq("app_id", dto.getAppId());
        wrapper.eq("task_name", dto.getTaskName());
        wrapper.ne("status", AppTaskStatusEnum.DELETED.value);
        boolean exists = iAppTaskService.exists(wrapper);
        if (exists) {
            throw new BizException("任务名称已存在");
        }
        AppTaskPo po = new AppTaskPo(dto);
        iAppTaskService.save(po);
        AppTaskBo bo = appTaskMapper.getAppTaskByTaskId(po.getTaskId());
        addTask(bo);
        return R.ok();
    }

    @Transactional(rollbackFor = Exception.class)
    public R<Object> updateAppTask(UpdateAppTaskDto dto) {
        checkAppTaskDto(dto);
        QueryWrapper<AppTaskPo> wrapper = new QueryWrapper<>();
        wrapper.eq("app_id", dto.getAppId());
        wrapper.eq("task_name", dto.getTaskName());
        wrapper.ne("status", AppTaskStatusEnum.DELETED.value);
        wrapper.ne("task_id", dto.getTaskId());
        boolean exists = iAppTaskService.exists(wrapper);
        if (exists) {
            throw new BizException("任务名称已存在");
        }
        AppTaskPo po = new AppTaskPo(dto);
        iAppTaskService.updateById(po);
        AppTaskBo bo = appTaskMapper.getAppTaskByTaskId(po.getTaskId());
        if (AppTaskStatusEnum.RUNNING.value.equals(bo.getStatus())) {
            addTask(bo);
        }
        return R.ok();
    }

    @Transactional(rollbackFor = Exception.class)
    public R<Object> enableAppTask(EnableAppTaskDto dto) {
        AppTaskPo po = new AppTaskPo();
        po.setTaskId(dto.getTaskId());
        po.setStatus(AppTaskStatusEnum.RUNNING.value);
        iAppTaskService.updateById(po);
        AppTaskBo bo = appTaskMapper.getAppTaskByTaskId(po.getTaskId());
        addTask(bo);
        return R.ok();
    }

    @Transactional(rollbackFor = Exception.class)
    public R<Object> stopAppTask(StopAppTaskDto dto) {
        AppTaskPo po = new AppTaskPo();
        po.setTaskId(dto.getTaskId());
        po.setStatus(AppTaskStatusEnum.STOPPED.value);
        iAppTaskService.updateById(po);
        removeTask(dto.getTaskId());
        return R.ok();
    }

    @Transactional(rollbackFor = Exception.class)
    public R<Object> deleteAppTask(DeleteAppTaskDto dto) {
        AppTaskPo po = new AppTaskPo();
        po.setTaskId(dto.getTaskId());
        po.setStatus(AppTaskStatusEnum.DELETED.value);
        iAppTaskService.updateById(po);
        removeTask(dto.getTaskId());
        return R.ok();
    }

    public R<Object> runAppTask(RunAppTaskDto dto) {
        AppTaskBo bo = appTaskMapper.getAppTaskByTaskId(dto.getTaskId());
        runTask(bo);
        return R.ok();
    }

    public R<PageVo<AppTaskLogPo>> listAppTaskLog(ListAppTaskLogDto dto) {
        QueryWrapper<AppTaskLogPo> wrapper = new QueryWrapper<>();
        if (StringUtils.hasText(dto.getId())) {
            wrapper.eq("id", dto.getId());
        }
        if (StringUtils.hasText(dto.getAppId())) {
            wrapper.eq("app_id", dto.getAppId());
        }
        if (StringUtils.hasText(dto.getAppName())) {
            wrapper.like("app_name", "%" + dto.getAppName() + "%");
        }
        if (StringUtils.hasText(dto.getTaskId())) {
            wrapper.eq("task_id", dto.getTaskId());
        }
        if (StringUtils.hasText(dto.getTaskName())) {
            wrapper.like("task_name", "%" + dto.getTaskName() + "%");
        }
        if (dto.getStatus() != null) {
            wrapper.eq("status", dto.getStatus());
        }
        if (StringUtils.hasText(dto.getRobotCode())) {
            wrapper.eq("robot_code", dto.getRobotCode());
        }
        if (StringUtils.hasText(dto.getRobotName())) {
            wrapper.like("robot_name", "%" + dto.getRobotName() + "%");
        }
        if (dto.getConversationType() != null) {
            wrapper.eq("conversation_type", dto.getConversationType());
        }
        if (StringUtils.hasText(dto.getOpenConversationId())) {
            wrapper.eq("open_conversation_id", dto.getOpenConversationId());
        }
        if (StringUtils.hasText(dto.getPhones())) {
            wrapper.like("phones", "%" + dto.getPhones() + "%");
        }
        if (StringUtils.hasText(dto.getUserIds())) {
            wrapper.like("user_ids", "%" + dto.getUserIds() + "%");
        }
        if (StringUtils.hasText(dto.getSftpId())) {
            wrapper.eq("sftp_id", dto.getSftpId());
        }
        if (StringUtils.hasText(dto.getHost())) {
            wrapper.like("host", "%" + dto.getHost() + "%");
        }
        if (dto.getStartTime() != null) {
            wrapper.ge("create_time", dto.getStartTime());
        }
        if (dto.getEndTime() != null) {
            wrapper.le("create_time", dto.getEndTime());
        }
        Page<AppTaskLogPo> page = new Page<>(dto.getCurrent(), dto.getPageSize());
        if (CollectionUtils.isEmpty(dto.getSort())) {
            page.addOrder(OrderItem.desc("create_time"));
        } else {
            List<OrderItem> items = dto.getSort().entrySet().stream()
                    .map(t -> t.getValue().equals(SortEnum.ascend) ? OrderItem.asc(t.getKey().value)
                            : OrderItem.desc(t.getKey().value)).toList();
            page.addOrder(items);
        }
        List<AppTaskLogPo> vos = iAppTaskLogService.list(page, wrapper);
        return R.ok(new PageVo<>(page.getTotal(), vos));
    }

    @Data
    public static class DirEntryBo {
        private String filename;
        private String longFilename;
        private Set<SftpClient.Attribute> flags;
        private int type;
        private int perms;
        private int uid;
        private int gid;
        private String owner;
        private String group;
        private long size;
        private FileTime accessTime;
        private FileTime createTime;
        private FileTime modifyTime;
        private List<AclEntry> acl;
        private Map<String, byte[]> extensions;

        public DirEntryBo(SftpClient.DirEntry t) {
            this.filename = t.getFilename();
            this.longFilename = t.getLongFilename();
            this.flags = t.getAttributes().getFlags();
            this.type = t.getAttributes().getType();
            this.perms = t.getAttributes().getPermissions();
            this.uid = t.getAttributes().getUserId();
            this.gid = t.getAttributes().getGroupId();
            this.owner = t.getAttributes().getOwner();
            this.group = t.getAttributes().getGroup();
            this.size = t.getAttributes().getSize();
            this.accessTime = t.getAttributes().getAccessTime();
            this.createTime = t.getAttributes().getCreateTime();
            this.modifyTime = t.getAttributes().getModifyTime();
            this.acl = t.getAttributes().getAcl();
            this.extensions = t.getAttributes().getExtensions();
        }
    }
}
